from typing import Any
from typing import Dict
from typing import Optional

from ..bliss_globals import setup_globals
from ..exafs.processor import ExafsProcessor
from ..import_utils import unavailable_module
from ..resources import resource_filename

try:
    from bliss.physics import units
except ImportError as ex:
    units = unavailable_module(ex)


class DemoExafsProcessor(ExafsProcessor):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("workflow", resource_filename("exafs", "exafs.ows"))
        defaults.setdefault("_scan_type", "any")
        counters = defaults.setdefault("_counters", dict())

        if self._HAS_BLISS:
            energy_unit = setup_globals.energy.unit or "eV"
            counters.setdefault(
                "any",
                {
                    "mu_name": "mu",
                    "energy_name": "energy",
                    "energy_unit": energy_unit,
                },
            )

        super().__init__(config=config, defaults=defaults)

    def _scan_type_from_scan(self, scan) -> Optional[str]:
        return "any"

    def _multi_xas_scan(self, scan) -> bool:
        return False

    def _multi_xas_subscan_size(self, scan) -> int:
        return 0

    def run(self, expo=0.003):
        e0 = 8800  # eV
        e1 = 9600  # eV
        step_size = 0.5  # eV
        intervals = int((e1 - e0) / step_size) - 1

        from_unit = "eV"
        to_unit = self.counters["energy_unit"]

        e0 = (e0 * units.ur(from_unit)).to(to_unit).magnitude
        e1 = (e1 * units.ur(from_unit)).to(to_unit).magnitude
        scan = setup_globals.ascan(
            setup_globals.energy, e0, e1, intervals, expo, setup_globals.mu, run=False
        )
        super().run(scan)


exafs_processor = DemoExafsProcessor()
exafs_processor.refresh_period = 0.5



import numpy

from ..id06.utils import get_positions

PI = numpy.pi


def test_get_positions():
    deg_step = 1
    rad_step = numpy.deg2rad(deg_step)

    pyfai_pos = get_positions(start=0, step=1, npts=360)
    theoretical_pos = numpy.concatenate(
        (
            numpy.arange(-PI / 2 + rad_step / 2, PI + rad_step / 2, rad_step),
            numpy.arange(-PI + rad_step / 2, -PI / 2 + rad_step / 2, rad_step),
        ),
        axis=0,
    )

    assert numpy.all(pyfai_pos - theoretical_pos < 1e-13)



import os
from typing import Any
from typing import Dict
from typing import Optional

from ..id12.converter import Id12Hdf5ToAsciiConverter
from . import EWOKS_RESULTS_DIR


class DemoId12Hdf5ToSpecConverter(Id12Hdf5ToAsciiConverter):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        root_dir = os.path.join(EWOKS_RESULTS_DIR, "id12", "inhouse")
        defaults.setdefault(
            "external_proposal_outdir", os.path.join(root_dir, "EXTERNAL")
        )
        defaults.setdefault(
            "inhouse_proposal_outdir", os.path.join(root_dir, "INHOUSE2")
        )
        defaults.setdefault("test_proposal_outdir", os.path.join(root_dir, "NOBACKUP"))

        super().__init__(config=config, defaults=defaults)

    def on_new_scan_metadata(self, scan) -> None:
        super().on_new_scan_metadata(scan)
        filename = self.output_filename(scan)
        dirname = os.path.dirname(filename)
        basename = os.path.basename(filename)
        scan_basename = f"scan{scan.scan_info['scan_nb']:03d}_{basename}"
        filename = os.path.join(dirname, scan_basename)
        print("Scan data will be saved in ASCII file", filename)


id12_converter = DemoId12Hdf5ToSpecConverter()



from blissoda.bliss_globals import setup_globals
from blissoda.demo.xrpd import xrpd_processor


def xrpd_demo(expo=0.2, npoints=10):
    xrpd_processor.enable(setup_globals.difflab6)
    try:
        pct(
            expo,
            setup_globals.difflab6,
            setup_globals.diode1,
            setup_globals.diode2,
        )
        setup_globals.loopscan(
            npoints,
            expo,
            setup_globals.difflab6,
            setup_globals.diode1,
            setup_globals.diode2,
        )
        xrpd_processor.integration_options["nbpt_azim"] = 360
        pct(
            expo,
            setup_globals.difflab6,
            setup_globals.diode1,
            setup_globals.diode2,
        )
        setup_globals.loopscan(
            npoints,
            expo,
            setup_globals.difflab6,
            setup_globals.diode1,
            setup_globals.diode2,
        )
    finally:
        xrpd_processor.integration_options.pop("nbpt_azim", None)
        xrpd_processor.disable()


def xrpd_demo_get_data():
    xrpd_processor.enable(setup_globals.difflab6)
    try:
        scan = setup_globals.loopscan(
            10,
            0.2,
            setup_globals.difflab6,
            setup_globals.diode1,
            setup_globals.diode2,
        )
    finally:
        xrpd_processor.disable()
    data_keys = xrpd_processor.get_data_keys(scan, "difflab6")
    return data_keys[0], xrpd_processor.get_data(scan, data_keys[0])


def pct(*args, **kw):
    s = setup_globals.ct(*args, **kw)
    xrpd_processor.on_new_scan(s)
    return s



"""User API for HDF5 conversion on the Bliss repl"""

import os
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ewoksjob.client import submit

from ..bliss_globals import current_session
from ..persistent.parameters import ParameterInfo
from ..processor import BaseProcessor
from ..processor import BlissScanType
from ..utils import directories


class Id14Hdf5ToSpecConverter(
    BaseProcessor,
    parameters=[
        ParameterInfo("workflow", category="workflows"),
        ParameterInfo("retry_timeout", category="data access"),
        ParameterInfo("queue", category="workflows", deprecated_names=["worker"]),
    ],
):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("queue", "celery")
        defaults.setdefault(
            "workflow", "/data/id14/inhouse/ewoks/resources/workflows/convert.json"
        )

        super().__init__(config=config, defaults=defaults)

    def on_new_scan_metadata(self, scan: BlissScanType) -> None:
        if not self.scan_requires_processing(scan):
            return
        kwargs = self.get_submit_arguments(scan)
        _ = submit(args=(self.workflow,), kwargs=kwargs, queue=self.queue)

    def _trigger_workflow_on_new_scan(self, scan: BlissScanType) -> None:
        self.on_new_scan_metadata(scan)

    def scan_requires_processing(self, scan: BlissScanType) -> bool:
        # TODO: select scan that needs processing
        return True

    def get_submit_arguments(self, scan: BlissScanType) -> dict:
        return {
            "inputs": self.get_inputs(scan),
            "outputs": [{"all": False}],
        }

    def get_inputs(self, scan: BlissScanType) -> List[dict]:
        task_identifier = "Hdf5ToSpec"

        filename = self.get_filename(scan)
        output_filename = self.workflow_destination(scan)
        scan_nb = scan.scan_info.get("scan_nb")

        inputs = [
            {
                "task_identifier": task_identifier,
                "name": "filename",
                "value": filename,
            },
            {
                "task_identifier": task_identifier,
                "name": "output_filename",
                "value": output_filename,
            },
            {
                "task_identifier": task_identifier,
                "name": "scan_numbers",
                "value": [scan_nb],
            },
        ]

        # Scan metadata published in id14.McaAcq.McaAcq.save
        calibration = scan.scan_info.get("instrument", dict()).get("calibration")
        if calibration:
            mca_calibration = calibration["a"], calibration["b"], 0
            inputs.append(
                {
                    "task_identifier": task_identifier,
                    "name": "mca_calibration",
                    "value": mca_calibration,
                }
            )

        return inputs

    def get_filename(self, scan: BlissScanType) -> str:
        filename = scan.scan_info.get("filename")
        if filename:
            return filename
        return current_session.scan_saving.filename

    def workflow_destination(self, scan: BlissScanType) -> str:
        filename = self.get_filename(scan)
        root = directories.get_processed_dir(filename)
        stem = os.path.splitext(os.path.basename(filename))[0]
        basename = f"{stem}.mca"
        return os.path.join(root, basename)

    def enable_slurm(self):
        self.queue = "slurm"

    def disable_slurm(self):
        self.queue = "celery"



import logging
from datetime import datetime
from typing import List
from typing import Literal
from typing import Optional
from typing import Union

from redis import Redis
from redis_om import JsonModel
from redis_om.model.model import NotFoundError

from ..bliss_globals import current_session
from ..import_utils import UnavailableObject
from ..import_utils import is_available
from ..persistent.ndarray import PersistentNdArray
from .compatibility import Field
from .compatibility import get_redis_db_url

logger = logging.getLogger(__name__)

XrpdFieldName = Literal["radial", "azimuthal", "intensity"]


def _create_database_proxy() -> Union[Redis, UnavailableObject]:
    if is_available(current_session):
        return Redis.from_url(get_redis_db_url())
    logger.warning("No Redis database without bliss")
    return UnavailableObject(ImportError())


class XrpdPlotInfo(JsonModel, frozen=True):
    scan_name: str
    lima_name: str
    radial_label: str
    azim_label: Optional[str]
    hdf5_url: Optional[str]
    timestamp: datetime = Field(default_factory=datetime.now)
    field_names: List[XrpdFieldName]

    @property
    def legend(self) -> str:
        return f"{self.scan_name} ({self.lima_name})"

    def _get_data_key(self, field_name: XrpdFieldName) -> str:
        return f"{self.scan_name}:{self.lima_name}:plot_data:{field_name}"

    def get_data_array(self, field_name: XrpdFieldName):
        if field_name not in self.field_names:
            raise KeyError(f"No field {field_name} in this PlotInfo")
        return PersistentNdArray(self._get_data_key(field_name))

    def delete_data_arrays(self):
        for field_name in self.field_names:
            PersistentNdArray(self._get_data_key(field_name)).remove()

    @classmethod
    def get(cls, pk):
        try:
            return super().get(pk)
        except NotFoundError:
            # Reraise NotFoundError to get more info in the error message
            raise KeyError(f"PlotInfo not found at {pk}@{cls._meta.database}")

    class Meta:
        database = _create_database_proxy()
